from ddeutil.observe.routes.workflow.schemas import (
    ReleaseLogCreate,
    WorkflowCreate,
)


def test_workflow_schema():
    # NOTE: This data return from ddeutil-workflow api.
    value = {
        "name": "wf-scheduling",
        "params": {"asat-dt": {"type": "datetime"}},
        "on": [
            {"cronjob": "*/3 * * * *", "timezone": "Asia/Bangkok"},
            {"cronjob": "* * * * *", "timezone": "Asia/Bangkok"},
        ],
        "jobs": {
            "condition-job": {
                "id": "condition-job",
                "stages": [
                    {"name": "Empty stage"},
                    {
                        "name": "Call-out",
                        "echo": "Hello ${{ params.asat-dt | fmt('%Y-%m-%d') }}",
                    },
                ],
            }
        },
    }
    rs = WorkflowCreate.model_validate(value)
    assert rs.name == "wf-scheduling"


def test_workflow_schema_log():
    value = {
        "release": "20240902093600",
        "logs": [
            {
                "run_id": "635351540020240902093554579053",
                "context": {
                    "name": "wf-scheduling",
                    "on": "*/3 * * * *",
                    "release": "2024-09-02 09:36:00+07:00",
                    "context": {
                        "params": {"asat-dt": "2024-09-02 09:36:00+07:00"},
                        "jobs": {
                            "condition-job": {
                                "matrix": {},
                                "stages": {
                                    "6708019737": {"outputs": {}},
                                    "0663452000": {"outputs": {}},
                                },
                            }
                        },
                    },
                    "parent_run_id": "635351540020240902093554579053",
                    "run_id": "635351540020240902093554579053",
                    "update": "2024-09-02 09:35:54.579053",
                },
            },
        ],
    }
    rs = ReleaseLogCreate.model_validate(value)
    assert rs.release == 20240902093600



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import datetime

from sqlalchemy.orm import Session
from sqlalchemy.sql import false

from ...crud import BaseCRUD
from ...utils import get_logger
from . import models, schemas

logger = get_logger("ddeutil.observe")


def get_workflow(session: Session, workflow_id: int) -> models.Workflows:
    return (
        session.query(models.Workflows)
        .filter(models.Workflows.id == workflow_id)
        .first()
    )


def get_workflow_by_name(session: Session, name: str) -> models.Workflows:
    return (
        session.query(models.Workflows)
        .filter(
            models.Workflows.name == name,
            models.Workflows.delete_flag == false(),
        )
        .first()
    )


def create_workflow(
    session: Session,
    workflow: schemas.WorkflowCreate,
) -> models.Workflows:
    db_workflow = models.Workflows(
        name=workflow.name,
        desc=workflow.desc,
        params=workflow.params,
        on=workflow.on,
        jobs=workflow.jobs,
        valid_start=datetime.now(),
        valid_end=datetime(2999, 12, 31),
    )
    session.add(db_workflow)
    session.commit()
    session.refresh(db_workflow)
    return db_workflow


def list_workflows(
    session: Session,
    skip: int = 0,
    limit: int = 1000,
) -> list[models.Workflows]:
    return (
        session.query(models.Workflows)
        .filter(models.Workflows.delete_flag == false())
        .offset(skip)
        .limit(limit)
        .all()
    )


def search_workflow(
    session: Session,
    search_text: str,
) -> list[models.Workflows]:
    if len(search_text) > 0:
        if not (search_text := search_text.strip().lower()):
            return []

        results = []
        for workflow in list_workflows(session=session):
            text: str = f"{workflow.name} {workflow.desc or ''}".lower()
            logger.debug(f"Getting text: {text} | Search {search_text}")
            if search_text in text:
                results.append(workflow)
        return results
    return list_workflows(session=session)


def get_release(
    session: Session,
    release: datetime,
) -> models.WorkflowReleases:
    return (
        session.query(models.WorkflowReleases)
        .filter(models.WorkflowReleases.release == release)
        .first()
    )


def create_release_log(
    session: Session,
    workflow_id: int,
    release_log: schemas.ReleaseLogCreate,
):
    db_release = models.WorkflowReleases(
        release=release_log.release,
        workflow_id=workflow_id,
    )
    session.add(db_release)
    session.commit()
    session.refresh(db_release)
    for log in release_log.logs:
        db_log = models.WorkflowLogs(
            run_id=log.run_id,
            context=log.context,
            release_id=db_release.id,
        )
        session.add(db_log)
        session.commit()
        session.refresh(db_log)
    return db_release


def get_log(session: Session, run_id: str) -> models.WorkflowLogs:
    return (
        session.query(models.WorkflowLogs)
        .filter(models.WorkflowLogs.run_id == run_id)
        .first()
    )


class WorkflowsCRUD(BaseCRUD):

    async def get_all(
        self,
        skip: int = 0,
        limit: int = 100,
    ) -> AsyncIterator[schemas.Workflow]:
        async for wf in models.Workflows.get_all(
            self.async_session,
            skip=skip,
            limit=limit,
            include_release=True,
        ):
            yield schemas.Workflow.model_validate(wf)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import logging
from functools import lru_cache

from .conf import config


@lru_cache
def get_logger(name: str):
    """Return logger with an input module name."""
    logger = logging.getLogger(name)
    formatter = logging.Formatter(
        fmt=(
            "%(asctime)s.%(msecs)03d (%(name)-10s, %(process)-5d, "
            "%(thread)-5d) [%(levelname)-7s] %(message)-120s "
            "(%(filename)s:%(lineno)s)"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    stream = logging.StreamHandler()
    stream.setFormatter(formatter)
    logger.addHandler(stream)
    logger.setLevel(
        logging.DEBUG if config.OBSERVE_LOG_DEBUG_MODE else logging.INFO
    )
    return logger



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Annotated, Optional

from fastapi import APIRouter, Depends, Header, Request
from fastapi.templating import Jinja2Templates
from sqlalchemy.orm import Session

from ...auth.deps import required_auth
from ...deps import get_session, get_templates
from ...utils import get_logger
from . import crud
from .schemas import (
    Workflow,
    Workflows,
)

logger = get_logger("ddeutil.observe")

workflow = APIRouter(
    prefix="/workflow",
    tags=["workflow"],
    dependencies=[Depends(required_auth)],
)


@workflow.get("/")
def read_workflows(
    request: Request,
    session: Session = Depends(get_session),
    templates: Jinja2Templates = Depends(get_templates),
):
    """Return all workflows."""
    workflows: list[Workflow] = Workflows.validate_python(
        crud.list_workflows(session)
    )
    return templates.TemplateResponse(
        request=request,
        name="workflow/workflow.html",
        context={
            "workflows": workflows,
            "search_text": "",
        },
    )


@workflow.get("/search")
def search_workflows(
    request: Request,
    search_text: str,
    hx_request: Annotated[Optional[str], Header(...)] = None,
    session: Session = Depends(get_session),
    templates: Jinja2Templates = Depends(get_templates),
):
    workflows: list[Workflow] = Workflows.validate_python(
        crud.search_workflow(session=session, search_text=search_text)
    )
    if hx_request:
        return templates.TemplateResponse(
            request=request,
            name="workflow/partials/workflow_results.html",
            context={"workflows": workflows},
        )
    return templates.TemplateResponse(
        request=request,
        name="workflow/workflow.html",
        context={
            "workflows": workflows,
            "search_text": search_text,
        },
    )


@workflow.get("/logs")
def read_logs(
    request: Request,
    hx_request: Annotated[Optional[str], Header(...)] = None,
    templates: Jinja2Templates = Depends(get_templates),
):
    """Return all workflows."""
    if hx_request:
        return templates.TemplateResponse(
            "workflow/partials/show_add_author_form.html", {"request": request}
        )
    return templates.TemplateResponse(request=request, name="workflow/log.html")



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime
from typing import Optional

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.types import DateTime, Integer, String

from ...db import Base, Col


class Tier(Base):
    __tablename__ = "tier"

    id: Mapped[int] = Col(Integer, primary_key=True)
    name: Mapped[str] = Col(String, nullable=False, unique=True)
    created_at: Mapped[datetime] = Col(
        DateTime(timezone=True),
        default=datetime.now,
        nullable=False,
    )
    updated_at: Mapped[Optional[datetime]] = Col(
        DateTime(timezone=True),
        default=None,
        nullable=True,
    )


class Group(Base):
    __tablename__ = "groups"

    id = Col(Integer, primary_key=True)
    name = Col(String, unique=True, nullable=False)
    member = Col(Integer, ForeignKey("users.id"))


class Role(Base):
    __tablename__ = "roles"

    id = Col(Integer, primary_key=True)
    name = Col(String, unique=True, nullable=False)


class Policy(Base):
    __tablename__ = "policies"

    id = Col(Integer, primary_key=True)
    name = Col(String, unique=True, nullable=False)



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import time
from contextlib import asynccontextmanager

from dotenv import load_dotenv
from fastapi import FastAPI, Request
from fastapi import status as st
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from sqlalchemy.exc import OperationalError

from .__about__ import __version__
from .auth import api_auth, auth
from .conf import config
from .db import sessionmanager
from .routes import api_router, workflow
from .utils import get_logger

load_dotenv()
logger = get_logger("ddeutil.observe")
sessionmanager.init(config.OBSERVE_SQLALCHEMY_DB_ASYNC_URL)


@asynccontextmanager
async def lifespan(_: FastAPI):
    yield
    if sessionmanager.is_opened():
        await sessionmanager.close()


app = FastAPI(
    titile="Observe Web",
    version=__version__,
    lifespan=lifespan,
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=[
        "http://localhost",
        "http://localhost:8080",
    ],
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.middleware("http")
async def add_process_time_header(request: Request, call_next):
    start_time = time.perf_counter()
    response = await call_next(request)
    process_time = time.perf_counter() - start_time
    response.headers["X-Process-Time"] = str(process_time)
    return response


@app.exception_handler(OperationalError)
async def sqlalchemy_exception_handler(_: Request, exc):
    return PlainTextResponse(
        str(exc.detail),
        status_code=st.HTTP_500_INTERNAL_SERVER_ERROR,
    )


# NOTE: Authentication
app.include_router(api_auth, prefix="/api/v1")
app.include_router(auth)

# NOTE: Any routers
app.include_router(api_router, prefix="/api/v1")
app.include_router(workflow)

app.mount("/static", StaticFiles(directory="static"), name="static")


@app.get("/")
async def home(request: Request):
    return RedirectResponse(
        request.url_for("login"), status_code=st.HTTP_303_SEE_OTHER
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from collections.abc import AsyncIterator
from typing import Any

from sqlalchemy import ForeignKey
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, relationship, selectinload
from sqlalchemy.sql.expression import select
from sqlalchemy.types import (
    JSON,
    Boolean,
    DateTime,
    Integer,
    String,
)
from typing_extensions import Self

from ...db import Base, Col


class Workflows(Base):
    __tablename__ = "workflows"

    id = Col(Integer, primary_key=True, index=True)
    name = Col(String, index=True)
    desc = Col(String)
    params: Mapped[dict[str, Any]] = Col(JSON)
    on: Mapped[dict[str, Any]] = Col(JSON)
    jobs: Mapped[dict[str, Any]] = Col(JSON)
    delete_flag = Col(Boolean, default=False)
    valid_start = Col(DateTime)
    valid_end = Col(DateTime)

    releases: Mapped[list[WorkflowReleases]] = relationship(
        "WorkflowReleases",
        back_populates="workflow",
    )

    @classmethod
    async def get_all(
        cls,
        session: AsyncSession,
        skip: int = 0,
        limit: int = 100,
        include_release: bool = False,
    ) -> AsyncIterator[Self]:
        stmt = select(cls)
        if include_release:
            stmt = stmt.options(selectinload(cls.releases))
        if skip > 0 and limit > 0:
            stmt = stmt.offset(skip).limit(limit)
        stream = await session.stream(stmt.order_by(cls.id))
        async for row in stream.scalars():
            yield row


class WorkflowReleases(Base):
    __tablename__ = "workflow_releases"

    id: Mapped[int] = Col(Integer, primary_key=True, index=True)
    release: Mapped[int] = Col(Integer, index=True)
    workflow_id: Mapped[int] = Col(Integer, ForeignKey("workflows.id"))

    workflow: Mapped[Workflows] = relationship(
        "Workflows", back_populates="releases"
    )
    logs: Mapped[list[WorkflowLogs]] = relationship(
        "WorkflowLogs",
        back_populates="release",
    )


class WorkflowLogs(Base):
    __tablename__ = "workflow_logs"

    run_id: Mapped[str] = Col(String, primary_key=True, index=True)
    context: Mapped[dict] = Col(JSON)
    release_id: Mapped[int] = Col(Integer, ForeignKey("workflow_releases.id"))

    release: Mapped[WorkflowReleases] = relationship(
        "WorkflowReleases",
        back_populates="logs",
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import pathlib
from collections.abc import AsyncIterator, Iterator

from fastapi import Request
from fastapi.templating import Jinja2Templates
from jinja2 import ChoiceLoader, FileSystemLoader
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from .db import SessionLocal, sessionmanager


def get_templates(request: Request) -> Jinja2Templates:
    """Dynamic multi-templating Jinja2 loader that support templates inside
    APIRouter.
    """
    choices: list[FileSystemLoader] = [FileSystemLoader("./templates")]
    if request.url.path != "/":
        route: str = request.url.path.strip("/").split("/")[0]
        route_path: pathlib.Path = (
            pathlib.Path(__file__).parent / f"routes/{route}/templates"
        )
        if route_path.exists():
            choices.insert(0, FileSystemLoader(route_path))
        else:
            auth_path: pathlib.Path = (
                pathlib.Path(__file__).parent / "auth/templates"
            )
            if auth_path.exists():
                choices.insert(0, FileSystemLoader(auth_path))

    return Jinja2Templates(
        directory="templates",
        loader=ChoiceLoader(choices),
    )


def get_session() -> Iterator[Session]:
    """Return the database local session instance."""
    session: Session = SessionLocal()
    try:
        yield session
    finally:
        session.close()


async def get_async_session() -> AsyncIterator[AsyncSession]:
    async with sessionmanager.session() as session:
        yield session



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime

from sqlalchemy import ForeignKey, text
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy.sql import select, true
from sqlalchemy.types import Boolean, DateTime, Integer, String
from typing_extensions import Self

from ...db import Base, Col


class Token(Base):
    __tablename__ = "tokens"

    id: Mapped[int] = Col(Integer, primary_key=True)
    access_token = Col(String(450), primary_key=True)
    refresh_token = Col(String(450), nullable=False)
    status = Col(Boolean, default=True)

    user_id = Col(Integer, ForeignKey("users.id"))

    expires_at: Mapped[datetime] = Col(DateTime)
    created_at: Mapped[datetime] = Col(
        DateTime,
        default=datetime.now,
        server_default=text("current_timestamp"),
    )
    updated_at: Mapped[datetime] = Col(
        DateTime,
        nullable=True,
        onupdate=datetime.now,
        server_default=text("current_timestamp"),
    )

    user = relationship(
        "User",
        back_populates="tokens",
    )

    @classmethod
    async def get_active_by_user(
        cls, session: AsyncSession, user_id: int
    ) -> list[Self]:
        return (
            (
                await session.execute(
                    select(cls).where(
                        cls.user_id == user_id,
                        cls.status == true(),
                    )
                )
            )
            .scalars()
            .all()
        )

    @classmethod
    async def get(cls, session: AsyncSession, token: str) -> Self | None:
        return (
            await session.execute(select(cls).where(cls.access_token == token))
        ).scalar_one_or_none()

    @classmethod
    async def get_by_refresh(
        cls, session: AsyncSession, token: str
    ) -> Self | None:
        return (
            (
                await session.execute(
                    select(cls)
                    .where(
                        cls.refresh_token == token,
                        cls.status == true(),
                    )
                    .order_by(cls.created_at)
                )
            )
            .scalars()
            .all()
        )

    @classmethod
    async def create(
        cls,
        session: AsyncSession,
        **kwargs,
    ):
        transaction = cls(**kwargs)
        session.add(transaction)
        await session.commit()
        await session.refresh(transaction)
        return transaction



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import os
import time
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

from sqlalchemy import MetaData, create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import (
    AsyncAttrs,
    AsyncConnection,
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, sessionmaker

from .utils import get_logger

logger = get_logger("ddeutil.observe")


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """Read more:
    - https://docs.sqlalchemy.org/en/20/dialects/sqlite.html -
        #foreign-key-support
    """
    cursor = dbapi_connection.cursor()
    settings: dict[str, Any] = {
        # "journal_mode": "WAL",
        "journal_mode": "OFF",
        "foreign_keys": "ON",
        "page_size": 4096,
        "cache_size": 10000,
        # "locking_mode": 'EXCLUSIVE',
        # "synchronous": "NORMAL",
        "synchronous": "OFF",
    }
    for k, v in settings.items():
        cursor.execute(f"PRAGMA {k} = {v};")
    cursor.close()


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(
    conn, cursor, statement, parameters, context, executemany
):
    conn.info.setdefault("query_start_time", []).append(time.time())
    logger.debug("Start Query: %s", statement)


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(
    conn, cursor, statement, parameters, context, executemany
):
    total = time.time() - conn.info["query_start_time"].pop(-1)
    logger.debug("Query Complete! Total Time: %f", total)


SQLALCHEMY_DATABASE_URL: str = os.getenv(
    "OBSERVE_SQLALCHEMY_DB_URL", "sqlite:///./observe.db"
)

engine = create_engine(
    SQLALCHEMY_DATABASE_URL,
    echo=False,
    connect_args={"check_same_thread": False},
)
SessionLocal = sessionmaker(
    autocommit=False, autoflush=False, future=True, bind=engine
)


class DBSessionManager:
    def __init__(self):
        self._engine: AsyncEngine | None = None
        self._sessionmaker: async_sessionmaker | None = None

    def init(self, host: str):
        self._engine = create_async_engine(
            host,
            echo=False,
            pool_pre_ping=True,
            connect_args={"check_same_thread": False},
        )
        self._sessionmaker = async_sessionmaker(
            autoflush=False,
            autocommit=False,
            future=True,
            expire_on_commit=False,
            bind=self._engine,
        )

    def is_opened(self) -> bool:
        return self._engine is not None

    async def close(self):
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")
        await self._engine.dispose()
        self._engine = None
        self._sessionmaker = None

    @asynccontextmanager
    async def connect(self) -> AsyncIterator[AsyncConnection]:
        if self._engine is None:
            raise Exception("DatabaseSessionManager is not initialized")

        async with self._engine.begin() as connection:
            try:
                yield connection
            except Exception:
                await connection.rollback()
                raise

    @asynccontextmanager
    async def session(self) -> AsyncIterator[AsyncSession]:
        if self._sessionmaker is None:
            raise Exception("DatabaseSessionManager is not initialized")

        session = self._sessionmaker()
        try:
            yield session
        except Exception:
            await session.rollback()
            raise
        finally:
            await session.close()

    @staticmethod
    async def create_all(connection: AsyncConnection):
        await connection.run_sync(Base.metadata.create_all)

    @staticmethod
    async def drop_all(connection: AsyncConnection):
        await connection.run_sync(Base.metadata.drop_all)


sessionmanager = DBSessionManager()


DB_INDEXES_NAMING_CONVENTION: dict[str, str] = {
    "ix": "%(column_0_label)s_idx",
    "uq": "%(table_name)s_%(column_0_name)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_check",
    "fk": "%(table_name)s_%(column_0_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}


# NOTE:
#       Attributes that are lazy-loading relationships, deferred columns or
#   expressions, or are being accessed in expiration scenarios can take
#   advantage of the AsyncAttrs mixin.
#   Read more: https://docs.sqlalchemy.org/en/20/orm/extensions/asyncio.html -
#       #preventing-implicit-io-when-using-asyncsession
#
class Base(AsyncAttrs, DeclarativeBase):
    __abstract__ = True

    metadata = MetaData(
        naming_convention=DB_INDEXES_NAMING_CONVENTION,
        # NOTE: In SQLite schema, the value should be `main` only because it
        #   does not implement with schema system.
        schema="main",
    )

    def __repr__(self) -> str:
        columns = ", ".join(
            [
                f"{k}={repr(v)}"
                for k, v in self.__dict__.items()
                if not k.startswith("_")
            ]
        )
        return f"<{self.__class__.__name__}({columns})>"


Col = mapped_column
Dtype = Mapped



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime
from typing import Annotated, Optional

from fastapi import Form
from pydantic import BaseModel, ConfigDict, EmailStr


class UserSchemaBase(BaseModel):
    username: str


class UserDetailSchema(UserSchemaBase):
    email: EmailStr
    fullname: Optional[str] = None


class UserSchema(UserDetailSchema):
    model_config = ConfigDict(from_attributes=True)

    id: int
    is_verified: bool
    is_active: bool
    is_superuser: bool
    created_at: datetime


class UserResetPassForm(UserSchemaBase):
    username: str
    old_password: str
    new_password: str


class UserCreateForm(UserDetailSchema):
    password: str

    # TODO: Move this validation step to client side.
    # @field_validator('password', mode='before')
    # def validate_strong_password(cls, value: str) -> str:
    #     if re.match(
    #         r'((?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?=.*\W).{8,64})',
    #         value,
    #     ):
    #         return value
    #     raise ValueError("Password does not strong")


class UserScopeForm(BaseModel):
    scopes_me: bool
    scopes_workflows: bool

    @classmethod
    def as_form(
        cls,
        scopes_me: Annotated[bool, Form()] = False,
        scopes_workflows: Annotated[bool, Form()] = False,
    ):
        return cls(
            scopes_me=scopes_me,
            scopes_workflows=scopes_workflows,
        )

    @property
    def scopes(self) -> list[str]:
        return [
            sc.split("_", maxsplit=1)[-1]
            for sc in self.__dict__
            if sc.startswith("scopes_") and self.__dict__[sc]
        ]



from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter


class WorkflowBase(BaseModel):
    """Base Workflow Pydantic model that does not include surrogate key column
    that create on the observe database.
    """

    name: str
    desc: Optional[str] = None
    params: dict[str, Any]
    on: list[dict[str, Any]]
    jobs: dict[str, Any]


class WorkflowCreate(WorkflowBase): ...


class Workflow(WorkflowBase):
    """Workflow Pydantic model that receive the Workflows model object
    from SQLAlchemy ORM.
    """

    model_config = ConfigDict(from_attributes=True)

    id: int
    delete_flag: bool
    valid_start: datetime
    valid_end: datetime


Workflows = TypeAdapter(list[Workflow])


class ReleaseBase(BaseModel):
    """Base Release Pydantic model that does not include surrogate key column
    that create on the observe database.
    """

    release: int


class Release(ReleaseBase):
    model_config = ConfigDict(from_attributes=True)

    id: int
    workflow_id: int


class LogBase(BaseModel):
    """Base Log Pydantic model that does not include surrogate key column
    that create on the observe database.
    """

    run_id: str
    context: dict[str, Any] = Field(default_factory=dict)


class LogCreate(LogBase): ...


class Log(LogBase):
    model_config = ConfigDict(from_attributes=True)

    release_id: int


class ReleaseLogCreate(ReleaseBase):
    logs: list[LogCreate] = Field(default_factory=list)


class ReleaseLog(ReleaseBase):
    model_config = ConfigDict(from_attributes=True)

    logs: list[Log]
    workflow_id: int



import os
import secrets

from ddeutil.core import str2bool

env = os.getenv


class BaseConfig:
    API_PREFIX: str = "api/v1/"

    OBSERVE_SQLALCHEMY_DB_ASYNC_URL: str = env(
        "OBSERVE_SQLALCHEMY_DB_ASYNC_URL",
        (
            "sqlite+aiosqlite://{DB_USER}:{DB_PASSWORD}@{DB_HOST}/{DB_NAME}"
        ).format(
            DB_USER=env("OBSERVE_DB_USER", ""),
            DB_PASSWORD=(
                f":{pwd}" if (pwd := env("OBSERVE_DB_PASSWORD")) else ""
            ),
            DB_HOST=env("OBSERVE_DB_HOST", ""),
            DB_NAME=env("OBSERVE_DB_NAME", "observe.db"),
        ),
    )
    OBSERVE_LOG_DEBUG_MODE: bool = str2bool(
        env("OBSERVE_LOG_DEBUG_MODE", "true")
    )

    # NOTE:
    #   token: 30 minutes = 30 minutes
    #   refresh: 60 minutes * 24 hours * 8 days = 8 days
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 30
    REFRESH_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8
    OBSERVE_SECRET_KEY: str = env(
        "OBSERVE_SECRET_KEY", secrets.token_urlsafe(32)
    )
    OBSERVE_REFRESH_SECRET_KEY: str = env(
        "OBSERVE_REFRESH_SECRET_KEY", secrets.token_urlsafe(32)
    )


config = BaseConfig



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime
from typing import Union

import jwt
from fastapi import HTTPException
from fastapi import status as st
from sqlalchemy import delete, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import false

from ..conf import config
from ..crud import BaseCRUD
from .models import Token, User
from .schemas import (
    TokenDataSchema,
    TokenRefreshCreate,
    UserCreateForm,
    UserResetPassForm,
    UserSchema,
)
from .securities import ALGORITHM, get_password_hash, verify_password


async def authenticate(
    session: AsyncSession,
    name: str,
    password: str,
) -> Union[User, bool]:
    if user := await User.get_by_username(session, username=name):
        return (
            user if verify_password(password, user.hashed_password) else False
        )
    return False


class TokenCRUD(BaseCRUD):

    async def retention_by_user(self, user_id: int):
        db_tokens: list[Token] = await Token.get_active_by_user(
            self.async_session, user_id
        )
        info: list[str] = []
        for record in db_tokens:
            if (datetime.now() - record.created_at).days > 1:
                info.append(record.id)

        if info:
            await self.async_session.execute(
                delete(Token).where(Token.id.in_(info))
            )
            await self.async_session.flush()
            await self.async_session.commit()

    async def update_logout(self, refresh_token: str):
        db_tokens = await Token.get_by_refresh(
            self.async_session, token=refresh_token
        )
        if db_tokens:
            rs = await self.async_session.execute(
                update(Token)
                .where(Token.get_by_refresh == refresh_token)
                .values(status=false())
                .returning(Token.access_token, Token.refresh_token)
            )
            await self.async_session.flush()
            await self.async_session.commit()
            return rs
        return []

    async def create(self, token: TokenRefreshCreate):
        db_token = Token(
            user_id=token.user_id,
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            status=token.status,
        )
        self.async_session.add(db_token)
        await self.async_session.flush()
        await self.async_session.commit()
        await self.async_session.refresh(db_token)
        return db_token


class UserCRUD(BaseCRUD):

    async def change_password(self, user: UserResetPassForm) -> UserSchema:
        db_user = await authenticate(
            self.async_session,
            name=user.username,
            password=user.old_password,
        )
        if db_user is None:
            raise HTTPException(
                status_code=st.HTTP_400_BAD_REQUEST,
                detail="User not found with old password",
            )

        encrypted_password = get_password_hash(user.new_password)
        db_user.password = encrypted_password
        await self.async_session.flush()
        await self.async_session.commit()
        await self.async_session.refresh(db_user)

        return UserSchema.model_validate(db_user)

    async def create_by_form(self, user: UserCreateForm) -> UserSchema:
        # NOTE: Validate by username value. By default, this will validate
        # from database with unique constraint.
        if await User.get_by_username(self.async_session, user.username):
            raise HTTPException(status_code=st.HTTP_409_CONFLICT)

        hashed_password = get_password_hash(user.password)
        db_user: User = User(
            email=user.email,
            username=user.username,
            hashed_password=hashed_password,
        )
        self.async_session.add(db_user)

        # `flush`, communicates a series of operations to the database
        # (insert, update, delete). The database maintains them as pending
        # operations in a transaction. The changes aren't persisted
        # permanently to disk, or visible to other transactions until the
        # database receives a COMMIT for the current transaction (which is
        # what session.commit() does).
        # ---
        # docs: https://stackoverflow.com/questions/4201455/ -
        #   sqlalchemy-whats-the-difference-between-flush-and-commit
        await self.async_session.flush()

        # # `commit`, commits (persists) those changes to the database.
        await self.async_session.commit()

        # NOTE: persisted some changes for an object to the database and
        # need to use this updated object within the same method.
        await self.async_session.refresh(db_user)
        return UserSchema.model_validate(db_user)


async def verify_token(
    token: str,
    session: AsyncSession,
) -> TokenDataSchema | None:
    if await Token.get(session, token=token):
        return None

    try:
        payload = jwt.decode(
            token, config.OBSERVE_REFRESH_SECRET_KEY, algorithms=[ALGORITHM]
        )
        if username := payload.get("sub"):
            return TokenDataSchema(username=username)
        return None
    except jwt.InvalidTokenError:
        return None


async def add_blacklist_token(token: str, session: AsyncSession) -> None:
    payload = jwt.decode(
        token, config.OBSERVE_SECRET_KEY, algorithms=[ALGORITHM]
    )
    expires_at = datetime.fromtimestamp(payload.get("exp"))
    await Token.create(
        session,
        **{"token": token, "expires_at": expires_at},
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Union

from pydantic import BaseModel, Field


class Token(BaseModel):
    access_token: str
    token_type: str = Field(default="Bearer")


class TokenRefresh(Token):
    refresh_token: str


class TokenRefreshCreate(TokenRefresh):
    user_id: int
    status: bool = Field(default=True)


class TokenRefreshForm(BaseModel):
    refresh_token: str


class TokenDataSchema(BaseModel):
    username: Union[str, None] = None
    scopes: list[str] = []



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Optional

import jwt
from fastapi import Depends, HTTPException, Security
from fastapi import status as st
from fastapi.security import SecurityScopes
from jwt.exceptions import InvalidTokenError
from pydantic import ValidationError
from sqlalchemy.ext.asyncio import AsyncSession

from ..conf import config
from ..deps import get_async_session
from .models import User
from .schemas import TokenDataSchema
from .securities import ALGORITHM, oauth2_scheme


async def get_current_access_token(
    token: Optional[str] = Depends(oauth2_scheme),
) -> Optional[str]:
    """Get the current access token."""
    return token


async def get_current_user(
    security_scopes: SecurityScopes,
    token: str = Depends(get_current_access_token),
    session: AsyncSession = Depends(get_async_session),
):
    """Get the current user async function that will."""
    if security_scopes.scopes:
        authenticate_value = f'Bearer scope="{security_scopes.scope_str}"'
    else:
        authenticate_value = "Bearer"

    credentials_exception = HTTPException(
        status_code=st.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(
            token, config.OBSERVE_SECRET_KEY, algorithms=[ALGORITHM]
        )

        if not (username := payload.get("sub")):
            raise credentials_exception

        token_scopes = payload.get("scopes", [])
        token_data = TokenDataSchema(scopes=token_scopes, username=username)
    except (InvalidTokenError, ValidationError):
        raise credentials_exception from None

    if not (
        user := await User.get_by_username(
            session, username=token_data.username
        )
    ):
        raise credentials_exception

    for scope in security_scopes.scopes:
        if scope not in token_data.scopes:
            raise HTTPException(
                status_code=st.HTTP_401_UNAUTHORIZED,
                detail="Not enough permissions",
                headers={"WWW-Authenticate": authenticate_value},
            )
    return user


async def get_current_active_user(
    current_user: User = Security(get_current_user, scopes=["me"]),
):
    if not current_user.is_active:
        raise HTTPException(
            status_code=st.HTTP_400_BAD_REQUEST,
            detail="Inactive user",
        )
    return current_user


async def required_auth(token: str = Depends(oauth2_scheme)):
    if not token:
        raise HTTPException(
            status_code=st.HTTP_307_TEMPORARY_REDIRECT,
            headers={"Location": "/auth/login"},
        )
    return True



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Annotated, Any

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi import status as st
from fastapi.security import OAuth2PasswordRequestForm
from sqlalchemy.ext.asyncio import AsyncSession

from ..deps import get_async_session
from ..utils import get_logger
from . import models
from .crud import TokenCRUD, authenticate, verify_token
from .deps import get_current_active_user
from .schemas import (
    Token,
    TokenRefresh,
    TokenRefreshCreate,
    UserSchema,
)
from .securities import (
    create_access_token,
    create_refresh_token,
)

logger = get_logger("ddeutil.observe")
auth = APIRouter(prefix="/auth", tags=["api", "auth"])


@auth.get("/user/{username}")
async def read_user(
    username: str,
    session: AsyncSession = Depends(get_async_session),
) -> UserSchema:
    return await models.User.get_by_username(session, username=username)


@auth.get("/user")
async def read_user_all(
    session: AsyncSession = Depends(get_async_session),
) -> list[UserSchema]:
    return await models.User.get_all(session)


@auth.post("/token")
async def token(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()],
    session: AsyncSession = Depends(get_async_session),
    service: TokenCRUD = Depends(TokenCRUD),
) -> TokenRefresh:
    if form_data.grant_type == "password":
        logger.debug("Authentication with user-password")
        user = await authenticate(
            session,
            name=form_data.username,
            password=form_data.password,
        )
    else:
        raise HTTPException(
            status_code=st.HTTP_406_NOT_ACCEPTABLE,
            detail=(
                f"grant type: {form_data.grant_type} does not support for this "
                f"application yet."
            ),
        )

    if not user:
        raise HTTPException(
            status_code=st.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    sub: dict[str, Any] = {
        "sub": user.username,
        "scopes": form_data.scopes,
    }
    access_token = create_access_token(subject=sub)
    refresh_token = create_refresh_token(subject=sub)
    return await service.create(
        token=TokenRefreshCreate(
            user_id=user.id,
            access_token=access_token,
            refresh_token=refresh_token,
        ),
    )


@auth.post("/refresh")
async def refresh(
    request: Request,
    session: AsyncSession = Depends(get_async_session),
) -> Token:
    refresh_token: str | None = request.cookies.get("refresh_token")
    if not refresh_token:
        raise HTTPException(
            status_code=st.HTTP_404_NOT_FOUND, detail="Refresh token missing."
        )

    if not (user_data := await verify_token(refresh_token, session)):
        raise HTTPException(
            status_code=st.HTTP_401_UNAUTHORIZED,
            detail="Invalid refresh token.",
        )

    new_access_token = create_access_token(subject={"sub": user_data.username})
    return {"access_token": new_access_token, "token_type": "Bearer"}


@auth.get("/token/me/", response_model=UserSchema)
async def read_user_me(
    current_user: Annotated[models.User, Depends(get_current_active_user)],
):
    """Get current active user from the current token."""
    return current_user



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from contextlib import asynccontextmanager

from fastapi import APIRouter, Depends, FastAPI, HTTPException
from fastapi import status as st
from sqlalchemy.orm.session import Session

from ...db import sessionmanager
from ...deps import get_session
from . import crud, models
from .crud import WorkflowsCRUD
from .schemas import ReleaseLog, ReleaseLogCreate, Workflow, WorkflowCreate


@asynccontextmanager
async def lifespan(_: FastAPI):
    """Lifespan for create workflow tables on target database."""
    async with sessionmanager.connect() as conn:
        # await conn.run_sync(models.Base.metadata.drop_all)
        await conn.run_sync(models.Base.metadata.create_all)
    yield


workflow = APIRouter(
    prefix="/workflow",
    tags=["api", "workflow"],
    lifespan=lifespan,
    responses={st.HTTP_404_NOT_FOUND: {"description": "Not found"}},
)


@workflow.get("/", response_model=list[Workflow])
async def read_all(
    skip: int = 0,
    limit: int = 100,
    service: WorkflowsCRUD = Depends(WorkflowsCRUD),
):
    return [wf async for wf in service.get_all(skip=skip, limit=limit)]


@workflow.post("/", response_model=Workflow)
def create_workflow(
    wf: WorkflowCreate, session: Session = Depends(get_session)
):
    db_workflow = crud.get_workflow_by_name(session, name=wf.name)
    if db_workflow:
        raise HTTPException(
            status_code=st.HTTP_302_FOUND,
            detail="Workflow already registered in observe database.",
        )
    return crud.create_workflow(session=session, workflow=wf)


@workflow.post("/{name}/release", response_model=ReleaseLog)
def create_workflow_release(
    name: str, rl: ReleaseLogCreate, session: Session = Depends(get_session)
):
    db_workflow = crud.get_workflow_by_name(session, name=name)
    if not db_workflow:
        raise HTTPException(
            status_code=st.HTTP_302_FOUND,
            detail="Workflow does not registered in observe database.",
        )
    return crud.create_release_log(
        session=session,
        workflow_id=db_workflow.id,
        release_log=rl,
    )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Annotated, Union

from fastapi import APIRouter, Depends, Form, Request
from fastapi import status as st
from fastapi.responses import HTMLResponse, Response
from fastapi.security import OAuth2PasswordRequestForm
from fastapi.templating import Jinja2Templates
from sqlalchemy.ext.asyncio import AsyncSession

from ..conf import config
from ..deps import get_async_session, get_templates
from .crud import TokenCRUD, UserCRUD, authenticate
from .deps import get_current_active_user
from .models import User
from .schemas import (
    TokenRefresh,
    UserCreateForm,
    UserResetPassForm,
    UserScopeForm,
)
from .securities import create_access_token, create_refresh_token

auth = APIRouter(prefix="/auth", tags=["auth", "frontend"])


@auth.get("/register")
def register(
    request: Request,
    template: Jinja2Templates = Depends(get_templates),
) -> HTMLResponse:
    return template.TemplateResponse(
        request=request,
        name="auth/authenticate.html",
        context={"content": "register"},
    )


@auth.post("/register")
async def register(
    response: Response,
    form_user: Annotated[UserCreateForm, Form()],
    service: UserCRUD = Depends(UserCRUD),
):
    """Register information of user to this application for able to access any
    routes.
    """
    await service.create_by_form(form_user)
    response.headers["HX-Redirect"] = "/auth/login/"
    response.status_code = st.HTTP_307_TEMPORARY_REDIRECT
    return {}


@auth.get("/login")
async def login(
    request: Request,
    templates: Jinja2Templates = Depends(get_templates),
) -> HTMLResponse:
    return templates.TemplateResponse(
        request=request,
        name="auth/authenticate.html",
        context={"request": request, "content": "login"},
    )


@auth.post("/login")
async def login(
    response: Response,
    session: AsyncSession = Depends(get_async_session),
    form_scopes: UserScopeForm = Depends(UserScopeForm.as_form),
    form_data: OAuth2PasswordRequestForm = Depends(OAuth2PasswordRequestForm),
):
    user: Union[User, bool] = await authenticate(
        session,
        name=form_data.username,
        password=form_data.password,
    )
    if not user:
        response.headers["HX-Redirect"] = "/auth/register"
        response.status_code = st.HTTP_404_NOT_FOUND
        return {}

    # NOTE: OAuth2 with scopes such as `["me", ...]`.
    access_token = create_access_token(
        subject={
            "sub": user.username,
            "scopes": form_scopes.scopes,
        },
    )
    refresh_token = create_refresh_token(
        subject={
            "sub": user.username,
            "scopes": form_scopes.scopes,
        },
    )

    # NOTE: Set cookies for access token and refresh token.
    response.set_cookie(
        key="access_token",
        value=f"Bearer {access_token}",
        expires=config.ACCESS_TOKEN_EXPIRE_MINUTES * 60,
        httponly=True,
    )
    response.set_cookie(
        key="refresh_token",
        value=refresh_token,
        httponly=True,
        secure=True,
        samesite="Lax",
        max_age=config.REFRESH_TOKEN_EXPIRE_MINUTES * 60,
    )
    response.headers["HX-Redirect"] = "/"
    response.status_code = st.HTTP_302_FOUND
    return {
        "access_token": access_token,
        "exp": config.ACCESS_TOKEN_EXPIRE_MINUTES * 60,
        "token_type": "Bearer",
    }


@auth.get("/change-password")
async def change_password(
    request: Request,
    templates: Jinja2Templates = Depends(get_templates),
) -> HTMLResponse:
    return templates.TemplateResponse(
        request=request,
        name="auth/authenticate.html",
        context={"request": request, "content": "change-password"},
    )


@auth.post("/change-password")
async def change_password(
    response: Response,
    form_data: Annotated[UserResetPassForm, Depends()],
    service: UserCRUD = Depends(UserCRUD),
):
    await service.change_password(form_data)

    response.headers["HX-Redirect"] = "/auth/login"
    response.status_code = st.HTTP_307_TEMPORARY_REDIRECT
    return {"message": "Password changed successfully"}


@auth.post("/logout")
async def logout(
    request: Request,
    response: Response,
    user: User = Depends(get_current_active_user),
    service: TokenCRUD = Depends(TokenCRUD),
):
    refresh_token = request.cookies.get("refresh_token")
    await service.retention_by_user(user.id)

    db_tokens = await service.update_logout(refresh_token)

    # NOTE: Delete cookies for access token and refresh token.
    response.delete_cookie(
        key="access_token",
        httponly=True,
    )
    response.delete_cookie(
        key="refresh_token",
        httponly=True,
    )

    response.headers["HX-Redirect"] = "/"
    response.status_code = st.HTTP_302_FOUND
    return {
        "message": "Logout Successfully",
        "logout": [
            TokenRefresh.model_validate(token).model_dump()
            for token in db_tokens
        ],
    }



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from fastapi import APIRouter

from .workflow.routes import workflow as workflow_api
from .workflow.views import workflow

api_router = APIRouter()
api_router.include_router(workflow_api)


@api_router.get("/", tags=["api"])
async def health():
    return {"message": "Observe Application Standby ..."}



"""
This file will contain script that will run before the app start to create the
super admin user.
"""

import asyncio

from .deps import get_async_session


async def create_admin(session): ...


async def main():
    async with get_async_session() as session:
        await create_admin(session)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from datetime import datetime, timedelta, timezone
from typing import Any, Optional, Union

import bcrypt
import jwt
from fastapi import HTTPException, Request
from fastapi import status as st
from fastapi.security import OAuth2PasswordBearer
from fastapi.security.utils import get_authorization_scheme_param

from ..conf import config

ALGORITHM: str = "HS256"


class OAuth2PasswordBearerCookie(OAuth2PasswordBearer):

    # IMPORTANT: it will raise Request does not exists when use
    #   `from __future__ import annotations` on above script file.
    async def __call__(self, request: Request) -> Optional[str]:

        # NOTE: if the header has authorization key, it will use this value with
        #   the first priority.
        if request.headers.get("Authorization"):
            return await super().__call__(request)

        # NOTE: get authorization key from the cookie.
        authorization: Optional[str] = request.cookies.get("access_token")
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            if self.auto_error:
                raise HTTPException(
                    status_code=st.HTTP_401_UNAUTHORIZED,
                    detail="Not authenticated",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            else:
                return None
        return param


oauth2_scheme = OAuth2PasswordBearerCookie(
    tokenUrl="api/v1/auth/token",
    scopes={
        "me": "Read information about the current user.",
        "workflows": "Read workflows and release logging.",
    },
    auto_error=False,
)


def create_access_token(
    subject: Union[str, dict[str, Any]],
    expires_delta: Union[timedelta, None] = None,
) -> str:
    if expires_delta:
        expire: datetime = datetime.now(timezone.utc) + expires_delta
    else:
        expire: datetime = datetime.now(timezone.utc) + timedelta(
            minutes=config.ACCESS_TOKEN_EXPIRE_MINUTES
        )

    if isinstance(subject, dict):
        to_encode = subject.copy()
        to_encode.update({"exp": expire})
    else:
        to_encode = {"exp": expire, "sub": str(subject)}

    return jwt.encode(to_encode, config.OBSERVE_SECRET_KEY, algorithm=ALGORITHM)


def create_refresh_token(
    subject: Union[str, dict[str, Any]],
    expires_delta: Union[timedelta, None] = None,
) -> str:
    if expires_delta:
        expire: datetime = datetime.now(timezone.utc) + expires_delta
    else:
        expire: datetime = datetime.now(timezone.utc) + timedelta(
            minutes=config.REFRESH_TOKEN_EXPIRE_MINUTES
        )

    if isinstance(subject, dict):
        to_encode = subject.copy()
        to_encode.update({"exp": expire})
    else:
        to_encode = {"exp": expire, "sub": str(subject)}

    return jwt.encode(
        to_encode, config.OBSERVE_REFRESH_SECRET_KEY, algorithm=ALGORITHM
    )


def verify_password(plain_password: str, hashed_password: str) -> bool:
    return bcrypt.checkpw(plain_password.encode(), hashed_password.encode())


def get_password_hash(password: str) -> str:
    return bcrypt.hashpw(password.encode(), bcrypt.gensalt()).decode()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from datetime import datetime
from uuid import UUID, uuid4

from sqlalchemy import text
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship, selectinload
from sqlalchemy.sql import select
from sqlalchemy.types import UUID, Boolean, DateTime, Integer, String
from typing_extensions import Self

from ...db import Base, Col, Dtype


class User(Base):
    __tablename__ = "users"

    id: Dtype[int] = Col(
        Integer,
        primary_key=True,
        index=True,
    )
    username: Dtype[str] = Col(
        String(64),
        unique=True,
        nullable=False,
        index=True,
    )
    fullname = Col(String(256), nullable=True)
    email: Dtype[str] = Col(String(128), nullable=True)
    hashed_password: Dtype[str] = Col(String, nullable=False)

    is_verified: Dtype[bool] = Col(Boolean, default=False)
    is_active: Dtype[bool] = Col(Boolean, default=True)
    is_superuser: Dtype[bool] = Col(Boolean, default=False)
    profile_image_url: Dtype[str] = Col(
        String, default="https://profileimageurl.com"
    )
    uuid: Dtype[UUID] = Col(
        UUID,
        default=uuid4,
        unique=True,
    )

    created_at: Dtype[datetime] = Col(DateTime, default=datetime.now)
    updated_at: Dtype[datetime] = Col(
        DateTime,
        nullable=True,
        onupdate=datetime.now,
        server_default=text("current_timestamp"),
    )
    deleted_at: Dtype[datetime] = Col(DateTime, default=datetime.now)

    tokens = relationship(
        "Token",
        back_populates="user",
        order_by="Token.created_at",
        cascade=(
            "save-update, merge, refresh-expire, expunge, delete, delete-orphan"
        ),
    )

    @classmethod
    async def create(
        cls, session: AsyncSession, user_id=None, **kwargs
    ) -> Self:
        if not user_id:
            user_id = uuid4().hex

        transaction = cls(id=user_id, **kwargs)
        session.add(transaction)
        await session.commit()
        await session.refresh(transaction)
        return transaction

    @classmethod
    async def get_by_username(
        cls,
        session: AsyncSession,
        username: str,
        *,
        include_tokens: bool = False,
    ) -> Self | None:
        stmt = select(cls).where(cls.username == username)
        if include_tokens:
            stmt = stmt.options(selectinload(cls.tokens))
        return (await session.execute(stmt)).scalar_one_or_none()

    @classmethod
    async def get_by_email(
        cls,
        session: AsyncSession,
        email: str,
    ) -> Self | None:
        try:
            return (
                (await session.execute(select(cls).where(cls.email == email)))
                .scalars()
                .first()
            )
        except NoResultFound:
            return None

    @classmethod
    async def get_all(cls, session: AsyncSession) -> list[Self]:
        return (await session.execute(select(cls))).scalars().all()



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession

from .deps import get_async_session


class BaseCRUD:
    def __init__(
        self,
        session: AsyncSession = Depends(get_async_session),
    ) -> None:
        self.async_session: AsyncSession = session
